from endesive.pdf import cms
from endesive.pdf.PyPDF2 import pdf, generic as po


class SignedData(cms.SignedData):
    def makepdf(self, prev, udct, algomd, zeros, cert, **params):
        catalog = prev.trailer["/Root"]
        size = prev.trailer["/Size"]
        pages = catalog["/Pages"].getObject()
        page0ref = prev.getPage(udct.get("sigpage", 0)).indirectRef

        self._objects = []
        while len(self._objects) < size - 1:
            self._objects.append(None)

        # if params['mode'] == 'timestamp':
        # deal with extensions
        if "/Extensions" not in catalog:
            extensions = po.DictionaryObject()
        else:
            extensions = catalog["/Extensions"]

        if "/ESIC" not in extensions:
            extensions.update(
                {
                    po.NameObject("/ESIC"): po.DictionaryObject(
                        {
                            po.NameObject("/BaseVersion"): po.NameObject("/1.7"),
                            po.NameObject("/ExtensionLevel"): po.NumberObject(1),
                        }
                    )
                }
            )
            catalog.update({po.NameObject("/Extensions"): extensions})
        else:
            esic = extensions["/ESIC"]
            major, minor = esic["/BaseVersion"].lstrip("/").split(".")
            if int(major) < 1 or int(minor) < 7:
                esic.update(
                    {
                        po.NameObject("/BaseVersion"): po.NameObject("/1.7"),
                        po.NameObject("/ExtensionLevel"): po.NumberObject(1),
                    }
                )

        # obj12 is the digital signature
        obj12, obj12ref = self._make_signature(
            Type=po.NameObject("/Sig"),
            # SubFilter=po.NameObject("/ETSI.CAdES.detached"),
            SubFilter=po.NameObject("/adbe.pkcs7.detached"),
            Contents=cms.UnencryptedBytes(zeros),
        )

        obj12.update(
            {
                po.NameObject("/Prop_Build"): pdf.DictionaryObject(
                    {
                        po.NameObject("/App"): pdf.DictionaryObject(
                            {
                                po.NameObject("/Name"): po.NameObject(
                                    "/" + udct.get("application", "endesive")
                                )
                            }
                        ),
                    }
                ),
            }
        )
        if params["mode"] == "timestamp":
            # obj12 is a timestamp this time
            obj12.update(
                {
                    po.NameObject("/Type"): po.NameObject("/DocTimeStamp"),
                    po.NameObject("/SubFilter"): po.NameObject("/ETSI.RFC3161"),
                    po.NameObject("/V"): po.NumberObject(0),
                }
            )
        else:
            obj12.update(
                {
                    po.NameObject("/Name"): po.createStringObject(udct["contact"]),
                    po.NameObject("/Location"): po.createStringObject(udct["location"]),
                    po.NameObject("/Reason"): po.createStringObject(udct["reason"]),
                }
            )
            if params.get("use_signingdate"):
                obj12.update(
                    {po.NameObject("/M"): po.createStringObject(udct["signingdate"])}
                )

        # obj13 is a combined AcroForm Sig field with Widget annotation
        new_13 = True
        # obj13 = po.DictionaryObject()
        if udct.get("signform", False):
            # Attaching signature to existing field in AcroForm
            if "/AcroForm" in catalog:
                form = catalog["/AcroForm"].getObject()
                if "/Fields" in form:
                    fields = form["/Fields"].getObject()
                    obj13ref = [
                        f
                        for f in fields
                        if f.getObject()["/T"] == udct.get("sigfield", "Signature1")
                    ][0]
                    obj13 = obj13ref.getObject()
                    self._objects[obj13ref.idnum - 1] = obj13
                    new_13 = False

        # box is coordinates of the annotation to fill
        box = udct.get("signaturebox", None)

        if new_13:
            obj13, obj13ref = self._make_sig_annotation(
                F=po.NumberObject(udct.get("sigflagsft", 132)),
                T=cms.EncodedString(udct.get("sigfield", "Signature1")),
                Vref=obj12ref,
                Pref=page0ref,
            )
        else:
            # original obj13 is a merged SigField/SigAnnot
            # Setting /V on the AcroForm field sets the signature
            # for the field
            obj13.update({po.NameObject("/V"): obj12ref})
            # fill the existing signature field annotation,
            # ignore any other location
            if "/Rect" in obj13:
                box = [float(f) for f in obj13["/Rect"]]

        # add an annotation if there is a field to fill
        if box is not None:
            self.addAnnotation(cert, udct, box, page0ref, obj13, obj13ref, new_13)

        if udct.get("sigandcertify", False) and "/Perms" not in catalog:
            obj10 = po.DictionaryObject()
            obj10ref = self._addObject(obj10)
            obj11 = po.DictionaryObject()
            obj11ref = self._addObject(obj11)
            obj14 = po.DictionaryObject()
            obj14ref = self._addObject(obj14)
            obj14.update({po.NameObject("/DocMDP"): obj12ref})
            obj10.update(
                {
                    po.NameObject("/Type"): po.NameObject("/TransformParams"),
                    po.NameObject("/P"): po.NumberObject(udct.get("sigflags", 3)),
                    po.NameObject("/V"): po.NameObject("/1.2"),
                }
            )
            obj11.update(
                {
                    po.NameObject("/Type"): po.NameObject("/SigRef"),
                    po.NameObject("/TransformMethod"): po.NameObject("/DocMDP"),
                    po.NameObject("/DigestMethod"): po.NameObject("/" + algomd.upper()),
                    po.NameObject("/TransformParams"): obj10ref,
                }
            )
            obj12[po.NameObject("/Reference")] = po.ArrayObject([obj11ref])
            catalog[po.NameObject("/Perms")] = obj14ref

        if "/AcroForm" in catalog:
            form = catalog["/AcroForm"].getObject()
            if "/Fields" in form:
                fields = form["/Fields"]
                old_field_names = [f.getObject()["/T"] for f in fields]
            else:
                fields = po.ArrayObject()
                old_field_names = []
            if udct.get("auto_sigfield", False) and obj13["/T"] in old_field_names:
                name_base = udct.get("sigfield", "Signature1")
                checklist = [
                    f[len(name_base) :]
                    for f in old_field_names
                    if f.startswith(name_base)
                ]
                for i in range(1, len(checklist) + 1):
                    suffix = "_{}".format(i)
                    if suffix in checklist:
                        continue

                    new_name = "{}{}".format(name_base, suffix)
                    obj13.update({po.NameObject("/T"): EncodedString(new_name)})
                    break

            old_flags = int(form.get("/SigFlags", 0))
            new_flags = int(form.get("/SigFlags", 0)) | udct.get("sigflags", 3)
            if new_13:
                fields.append(obj13ref)
                form.update(
                    {
                        po.NameObject("/Fields"): fields,
                        po.NameObject("/SigFlags"): po.NumberObject(new_flags),
                    }
                )
            elif new_flags > old_flags:
                form.update({po.NameObject("/SigFlags"): po.NumberObject(new_flags)})
            formref = catalog.raw_get("/AcroForm")
            if isinstance(formref, po.IndirectObject):
                self._objects[formref.idnum - 1] = form
                form = formref
        else:
            form = po.DictionaryObject()
            form.update(
                {
                    po.NameObject("/Fields"): po.ArrayObject([obj13ref]),
                    po.NameObject("/SigFlags"): po.NumberObject(
                        udct.get("sigflags", 3)
                    ),
                }
            )
        catalog[po.NameObject("/AcroForm")] = form

        if "/Metadata" in catalog:
            catalog[po.NameObject("/Metadata")] = catalog.raw_get("/Metadata")

        x_root = prev.trailer.raw_get("/Root")
        self._objects[x_root.idnum - 1] = catalog
        self.x_root = po.IndirectObject(x_root.idnum, 0, self)
        self.x_info = prev.trailer.get("/Info")


def sign(
    datau,
    udct,
    key,
    cert,
    othercerts,
    algomd="sha1",
    hsm=None,
    timestampurl=None,
    timestampcredentials=None,
    timestamp_req_options=None,
    ocspurl=None,
    ocspissuer=None,
):
    cls = SignedData()
    return cls.sign(
        datau,
        udct,
        key,
        cert,
        othercerts,
        algomd,
        hsm,
        timestampurl,
        timestampcredentials,
        timestamp_req_options,
        "sign",
        ocspurl,
        ocspissuer,
    )



from endesive import pdf, signer

import datetime
import hashlib
import base64
import io


import requests
import jwt


from . import psuhsm
from . import psucms
from . import exception


class PSUSigner:
    def __init__(
        self,
        code,
        secret,
        agent_key,
        jwt_secret=None,
        api_url=None,
    ):
        self.jwt_secret = jwt_secret
        self.agent_key = agent_key
        self.code = code
        self.secret = secret
        self.api_url = api_url
        self.hsm = None

        if jwt_secret:
            self.hsm = psuhsm.JWTSecretPSUHSM(
                code=code,
                secret=secret,
                jwt_secret=jwt_secret,
                agent_key=agent_key,
                api_url=api_url,
            )
        else:
            self.hsm = psuhsm.SecretPSUHSM(
                code=code, secret=secret, agent_key=agent_key, api_url=api_url
            )

        # function stuff
        self.original_signer = signer.sign
        signer.sign = self.sign_cms
        self.logs_api_url = f"{self.api_url}/sign/logs"

    def sign_byte(self, datau, dct, **kwargs):
        self.hsm.set_sign_parameters(**kwargs)
        datas = psucms.sign(datau, dct, None, None, [], "sha256", self.hsm)

        output = io.BytesIO()

        output.write(datau)
        output.write(datas)
        output.seek(0)

        return output

    def sign_file(self, input_name, dct, output_name, **kwargs):
        datau = b""
        with open(input_name, "rb") as fp:
            datau = fp.read()

        if not datau:
            raise exception.PSUSignException("No Input Data")

        self.hsm.set_sign_parameters(**kwargs)
        datas = psucms.sign(datau, dct, None, None, [], "sha256", self.hsm)

        with open(output_name, "wb") as fp:
            fp.write(datau)
            fp.write(datas)

    def search(self, **kwargs):
        """
        https://ds-dev.psu.ac.th/api/v1/sign/logs?ref1=ref1&ref2=ref2&max=10&offset=0&sort=actionDateTime&order=desc

        params

        certEntryCode: c1
        ref1: ref1 (exact match)
        ref2: ref2 (partial match)
        fromActionDateTime: 2023-06-05T11:00:00 (must apply when filter by action datetime)
        toActionDateTime: 2023-06-05T12:00:00 (optional when filter by action datetime)
        max: 10 (maximum result)
        offset: 0 (result offset for pagination)
        sort: actionDateTime [id, ref1, ref2, actionDateTime]
        order: desc [asc, desc]
        """
        headers = {"X-Agent-Key": self.agent_key}
        response = requests.get(self.logs_api_url, headers=headers, params=kwargs)
        return response.json()

    def sign_cms(
        self,
        datau,
        key,
        cert,
        othercerts,
        hashalgo,
        attrs=True,
        signed_value=None,
        hsm=None,
        pss=False,
        timestampurl=None,
        timestampcredentials=None,
        timestamp_req_options=None,
        ocspurl=None,
        ocspissuer=None,
    ):
        # print(
        #     datau,
        #     key,
        #     cert,
        #     othercerts,
        #     hashalgo,
        #     attrs,
        #     signed_value,
        #     hsm,
        #     pss,
        #     timestampurl,
        #     timestampcredentials,
        #     timestamp_req_options,
        #     ocspurl,
        #     ocspissuer,
        # )

        if hsm and isinstance(hsm, psuhsm.PSUHSM):
            if signed_value is None:
                signed_value = getattr(hashlib, hashalgo)(datau).digest()

            return hsm.sign(None, signed_value, "HA256")

        return self.original_signer(
            datau,
            key,
            cert,
            othercerts,
            hashalgo,
            attrs,
            signed_value,
            hsm,
            pss,
            timestampurl,
            timestampcredentials,
            timestamp_req_options,
            ocspurl,
            ocspissuer,
        )



import datetime
import jwt
import base64
import requests

from endesive import hsm, pdf, signer

from . import exception


class PSUHSM(hsm.BaseHSM):
    def __init__(self, code, secret, agent_key, api_url=None, timeout=120):
        self.api_url = "https://ds.psu.ac.th/api/v1"
        if api_url:
            self.api_url = api_url

        self.sign_api_url = f"{self.api_url}/sign/withPdfSignature"
        self.log_api_url = f"{self.api_url}/sign/logs"

        self.code = code
        self.secret = secret
        self.agent_key = agent_key
        self.sign_parameters = {}
        self.sign_timeout = timeout

    def set_sign_parameters(self, **kwargs):
        self.sign_parameters = kwargs

    def certificate(self):
        return None, None

    def get_headers(self):
        headers = {"X-Agent-Key": self.agent_key}

        return headers

    def get_payload(self, **kwargs):
        kwargs.update(self.sign_parameters)
        return kwargs

    def sign(self, keyid, data, mech):
        headers = self.get_headers()
        payload = self.get_payload(data)

        # print("headers:", headers)
        # print("payload:", payload)
        response = requests.post(
            self.sign_api_url, json=payload, headers=headers, timeout=self.timeout
        )

        if response.status_code != 200:
            raise exception.PSUSignException(response.json())

        json_data = response.json()
        status_code = json_data.get("statusCode")
        if status_code != 1:
            raise exception.PSUSignException(response.json())

        sign_data = json_data.get("signData")

        return base64.b64decode(sign_data)


class SecretPSUHSM(PSUHSM):
    def __init__(self, code, secret, agent_key, api_url=None):
        super().__init__(code, secret, agent_key, api_url)

    def get_payload(self, signed_value):
        pdf_hash = base64.b64encode(signed_value)

        data = {"code": self.code, "secret": self.secret, "pdfHash": pdf_hash.decode()}

        data.update(self.sign_parameters)
        return data


class JWTSecretPSUHSM(SecretPSUHSM):
    def __init__(self, code, secret, agent_key, jwt_secret=None, api_url=None):
        super().__init__(code, secret, agent_key, api_url)
        self.jwt_secret = jwt_secret

    def get_payload(self, signed_value):
        data = super().get_payload(signed_value)
        # print("jwt data", data)
        encoded = jwt.encode(data, self.jwt_secret, algorithm="HS256")

        data = {"payload": encoded}

        return data
